@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2020, The PurpleI2P Project
+* Copyright (c) 2013-2024, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -12,8 +12,10 @@
 #include <inttypes.h>
 #include <string>
 #include <memory>
+#include <mutex>
 #include <thread>
 #include <map>
+#include <unordered_map>
 #include <boost/asio.hpp>
 #include "util.h"
 #include "Destination.h"
@@ -58,9 +60,19 @@ namespace client
 		eI2CPMessageStatusAccepted = 1,
 		eI2CPMessageStatusGuaranteedSuccess = 4,
 		eI2CPMessageStatusGuaranteedFailure = 5,
+		eI2CPMessageStatusNoLocalTunnels = 16,
 		eI2CPMessageStatusNoLeaseSet = 21
 	};
 
+	enum I2CPSessionStatus
+	{
+		eI2CPSessionStatusDestroyed = 0,
+		eI2CPSessionStatusCreated = 1,
+		eI2CPSessionStatusUpdated = 2,
+		eI2CPSessionStatusInvalid = 3,
+		eI2CPSessionStatusRefused = 4
+	};
+
 	// params
 	const char I2CP_PARAM_MESSAGE_RELIABILITY[] = "i2cp.messageReliability";
 
@@ -69,8 +81,9 @@ namespace client
 	{
 		public:
 
-			I2CPDestination (boost::asio::io_service& service, std::shared_ptr<I2CPSession> owner,
-				std::shared_ptr<const i2p::data::IdentityEx> identity, bool isPublic, const std::map<std::string, std::string>& params);
+			I2CPDestination (boost::asio::io_context& service, std::shared_ptr<I2CPSession> owner,
+				std::shared_ptr<const i2p::data::IdentityEx> identity, bool isPublic, bool isSameThread, 
+			    const std::map<std::string, std::string>& params);
 			~I2CPDestination () {};
 
 			void Stop ();
@@ -81,7 +94,8 @@ namespace client
 			void LeaseSetCreated (const uint8_t * buf, size_t len); // called from I2CPSession
 			void LeaseSet2Created (uint8_t storeType, const uint8_t * buf, size_t len); // called from I2CPSession
 			void SendMsgTo (const uint8_t * payload, size_t len, const i2p::data::IdentHash& ident, uint32_t nonce); // called from I2CPSession
-
+			bool SendMsg (const uint8_t * payload, size_t len, std::shared_ptr<i2p::garlic::GarlicRoutingSession> remoteSession, uint32_t nonce);
+			
 			// implements LocalDestination
 			bool Decrypt (const uint8_t * encrypted, uint8_t * data, i2p::data::CryptoKeyType preferredCrypto) const;
 			bool SupportsEncryptionType (i2p::data::CryptoKeyType keyType) const;
@@ -90,6 +104,7 @@ namespace client
 
 		protected:
 
+			void CleanupDestination ();
 			// I2CP
 			void HandleDataMessage (const uint8_t * buf, size_t len);
 			void CreateNewLeaseSet (const std::vector<std::shared_ptr<i2p::tunnel::InboundTunnel> >& tunnels);
@@ -99,7 +114,9 @@ namespace client
 			std::shared_ptr<I2CPDestination> GetSharedFromThis ()
 			{ return std::static_pointer_cast<I2CPDestination>(shared_from_this ()); }
 			bool SendMsg (std::shared_ptr<I2NPMessage> msg, std::shared_ptr<const i2p::data::LeaseSet> remote);
-
+			bool SendMsg (std::shared_ptr<I2NPMessage> garlic, 
+				std::shared_ptr<i2p::tunnel::OutboundTunnel> outboundTunnel, std::shared_ptr<const i2p::data::Lease> remoteLease);
+			
 			void PostCreateNewLeaseSet (std::vector<std::shared_ptr<i2p::tunnel::InboundTunnel> > tunnels);
 
 		private:
@@ -111,8 +128,9 @@ namespace client
 			std::shared_ptr<i2p::crypto::ECIESX25519AEADRatchetDecryptor> m_ECIESx25519Decryptor;
 			uint8_t m_ECIESx25519PrivateKey[32];
 			uint64_t m_LeaseSetExpirationTime;
-			bool m_IsCreatingLeaseSet;
+			bool m_IsCreatingLeaseSet, m_IsSameThread;
 			boost::asio::deadline_timer m_LeaseSetCreationTimer;
+			i2p::util::MemoryPoolMt<I2NPMessageBuffer<I2NP_MAX_MESSAGE_SIZE> > m_I2NPMsgsPool;
 	};
 
 	class RunnableI2CPDestination: private i2p::util::RunnableService, public I2CPDestination
@@ -132,13 +150,7 @@ namespace client
 	{
 		public:
 
-#ifdef ANDROID
-			typedef boost::asio::local::stream_protocol proto;
-#else
-			typedef boost::asio::ip::tcp proto;
-#endif
-
-			I2CPSession (I2CPServer& owner, std::shared_ptr<proto::socket> socket);
+			I2CPSession (I2CPServer& owner, std::shared_ptr<boost::asio::ip::tcp::socket> socket);
 
 			~I2CPSession ();
 
@@ -151,6 +163,8 @@ namespace client
 			void SendI2CPMessage (uint8_t type, const uint8_t * payload, size_t len);
 			void SendMessagePayloadMessage (const uint8_t * payload, size_t len);
 			void SendMessageStatusMessage (uint32_t nonce, I2CPMessageStatus status);
+			void AddRoutingSession (const i2p::data::IdentHash& signingKey, std::shared_ptr<i2p::garlic::GarlicRoutingSession> remoteSession);
+			void CleanupRoutingSessions ();
 
 			// message handlers
 			void GetDateMessageHandler (const uint8_t * buf, size_t len);
@@ -180,17 +194,19 @@ namespace client
 			std::string ExtractString (const uint8_t * buf, size_t len);
 			size_t PutString (uint8_t * buf, size_t len, const std::string& str);
 			void ExtractMapping (const uint8_t * buf, size_t len, std::map<std::string, std::string>& mapping);
-			void SendSessionStatusMessage (uint8_t status);
+			void SendSessionStatusMessage (I2CPSessionStatus status);
 			void SendHostReplyMessage (uint32_t requestID, std::shared_ptr<const i2p::data::IdentityEx> identity);
-
+			
 		private:
 
 			I2CPServer& m_Owner;
-			std::shared_ptr<proto::socket> m_Socket;
+			std::shared_ptr<boost::asio::ip::tcp::socket> m_Socket;
 			uint8_t m_Header[I2CP_HEADER_SIZE], m_Payload[I2CP_MAX_MESSAGE_LENGTH];
 			size_t m_PayloadLen;
 
 			std::shared_ptr<I2CPDestination> m_Destination;
+			std::mutex m_RoutingSessionsMutex;
+			std::unordered_map<i2p::data::IdentHash, std::shared_ptr<i2p::garlic::GarlicRoutingSession> > m_RoutingSessions; // signing key->session
 			uint16_t m_SessionID;
 			uint32_t m_MessageID;
 			bool m_IsSendAccepted;
@@ -206,24 +222,23 @@ namespace client
 	{
 		public:
 
-			I2CPServer (const std::string& interface, int port, bool isSingleThread);
+			I2CPServer (const std::string& interface, uint16_t port, bool isSingleThread);
 			~I2CPServer ();
 
 			void Start ();
 			void Stop ();
-			boost::asio::io_service& GetService () { return GetIOService (); };
+			auto& GetService () { return GetIOService (); };
 			bool IsSingleThread () const { return m_IsSingleThread; };
 
 			bool InsertSession (std::shared_ptr<I2CPSession> session);
 			void RemoveSession (uint16_t sessionID);
+			std::shared_ptr<I2CPSession> FindSessionByIdentHash (const i2p::data::IdentHash& ident) const;
 
 		private:
 
-			void Run ();
-
 			void Accept ();
 
-			void HandleAccept(const boost::system::error_code& ecode, std::shared_ptr<I2CPSession::proto::socket> socket);
+			void HandleAccept(const boost::system::error_code& ecode, std::shared_ptr<boost::asio::ip::tcp::socket> socket);
 
 		private:
 
@@ -231,7 +246,7 @@ namespace client
 			I2CPMessageHandler m_MessagesHandlers[256];
 			std::map<uint16_t, std::shared_ptr<I2CPSession> > m_Sessions;
 
-			I2CPSession::proto::acceptor m_Acceptor;
+			boost::asio::ip::tcp::acceptor m_Acceptor;
 
 		public:
 
